@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2001 Mikael Hallendal <micke@codefactory.se>
+ * Copyright (C) 2001-2002 Mikael Hallendal <micke@codefactory.se>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -35,24 +35,42 @@
 
 #include "yelp-html.h"
 #include "yelp-marshal.h"
+#include "yelp-reader.h"
 #include "yelp-scrollkeeper.h"
 #include "yelp-util.h"
+#include "yelp-uri.h"
 #include "yelp-view-content.h"
 
 #define d(x)
 
-static void yvc_init                      (YelpViewContent         *html);
-static void yvc_class_init                (YelpViewContentClass    *klass);
-static void yvc_tree_selection_changed_cb (GtkTreeSelection        *selection,
-					   YelpViewContent         *content);
-static void yvc_html_uri_selected_cb      (YelpHtml                *html,
-					   YelpURI                 *uri,
-					   gboolean                 handled,
-					   YelpViewContent         *view);
-
+static void content_init                      (YelpViewContent      *html);
+static void content_class_init                (YelpViewContentClass *klass);
+static void content_tree_selection_changed_cb (GtkTreeSelection     *selection,
+				   	       YelpViewContent      *content);
+static void content_html_uri_selected_cb      (YelpHtml             *html,
+					       YelpURI              *uri,
+					       gboolean              handled,
+					       YelpViewContent      *view);
+static void content_html_title_changed_cb     (YelpHtml             *html,
+					       const gchar          *title,
+					       YelpViewContent      *view);
+static void content_reader_data_cb            (YelpReader           *reader,
+					       const gchar          *data,
+					       gint                  len,
+					       YelpViewContent      *view);
+static void content_reader_finished_cb        (YelpReader           *reader,
+					       YelpURI              *uri,
+					       YelpViewContent      *view);
+static void content_reader_error_cb           (YelpReader           *reader,
+					       GError               *error,
+					       YelpViewContent      *view);
+static GtkTreePath *
+content_find_path_from_uri                    (GtkTreeModel         *model,
+					       YelpURI              *uri);
 
 enum {
 	URI_SELECTED,
+	TITLE_CHANGED,
 	LAST_SIGNAL
 };
 
@@ -65,10 +83,15 @@ struct _YelpViewContentPriv {
 	GtkTreeStore *tree_store;
 	GNode        *doc_tree;
 
+	YelpReader   *reader;
+
 	/* Html view */
 	YelpHtml     *html_view;
+	GtkWidget    *html_widget;
 
-	gchar        *current_docpath;
+	YelpURI      *current_uri;
+	
+	gboolean      first;
 };
 
 GType
@@ -83,12 +106,12 @@ yelp_view_content_get_type (void)
                                 sizeof (YelpViewContentClass),
                                 NULL,
                                 NULL,
-                                (GClassInitFunc) yvc_class_init,
+                                (GClassInitFunc) content_class_init,
                                 NULL,
                                 NULL,
                                 sizeof (YelpViewContent),
                                 0,
-                                (GInstanceInitFunc) yvc_init,
+                                (GInstanceInitFunc) content_init,
                         };
                 
                 view_type = g_type_register_static (GTK_TYPE_HPANED,
@@ -101,7 +124,7 @@ yelp_view_content_get_type (void)
 
 
 static void
-yvc_init (YelpViewContent *view)
+content_init (YelpViewContent *view)
 {
 	YelpViewContentPriv *priv;
 	
@@ -117,15 +140,33 @@ yvc_init (YelpViewContent *view)
 				 GTK_TREE_MODEL (priv->tree_store));
 	
 	priv->html_view       = yelp_html_new ();
-	priv->current_docpath = g_strdup ("");
+	priv->html_widget     = yelp_html_get_widget (priv->html_view);
+/* 	priv->current_docpath = g_strdup (""); */
+	priv->current_uri     = NULL;
+	priv->first         = FALSE;
 	
 	g_signal_connect (priv->html_view, "uri_selected",
-			  G_CALLBACK (yvc_html_uri_selected_cb), 
+			  G_CALLBACK (content_html_uri_selected_cb), 
+			  view);
+	g_signal_connect (priv->html_view, "title_changed",
+			  G_CALLBACK (content_html_title_changed_cb),
+			  view);
+
+	priv->reader      = yelp_reader_new ();
+	
+	g_signal_connect (G_OBJECT (priv->reader), "data",
+			  G_CALLBACK (content_reader_data_cb),
+			  view);
+	g_signal_connect (G_OBJECT (priv->reader), "finished",
+			  G_CALLBACK (content_reader_finished_cb),
+			  view);
+	g_signal_connect (G_OBJECT (priv->reader), "error",
+			  G_CALLBACK (content_reader_error_cb),
 			  view);
 }
 
 static void
-yvc_class_init (YelpViewContentClass *klass)
+content_class_init (YelpViewContentClass *klass)
 {
 	signals[URI_SELECTED] = 
 		g_signal_new ("uri_selected",
@@ -137,10 +178,21 @@ yvc_class_init (YelpViewContentClass *klass)
 			      yelp_marshal_VOID__POINTER_BOOLEAN,
 			      G_TYPE_NONE,
 			      2, G_TYPE_POINTER, G_TYPE_BOOLEAN);
+
+	signals[TITLE_CHANGED] = 
+		g_signal_new ("title_changed",
+			      G_TYPE_FROM_CLASS (klass),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (YelpViewContentClass,
+					       title_changed),
+			      NULL, NULL,
+			      yelp_marshal_VOID__STRING,
+			      G_TYPE_NONE,
+			      1, G_TYPE_STRING);
 }
 
 static void
-yvc_html_uri_selected_cb (YelpHtml        *html,
+content_html_uri_selected_cb (YelpHtml        *html,
 			  YelpURI         *uri,
 			  gboolean         handled,
 			  YelpViewContent *view)
@@ -154,8 +206,16 @@ yvc_html_uri_selected_cb (YelpHtml        *html,
 }
 
 static void
-yvc_tree_selection_changed_cb (GtkTreeSelection *selection, 
-			       YelpViewContent  *content)
+content_html_title_changed_cb (YelpHtml        *html,
+			       const gchar     *title,
+			       YelpViewContent *view)
+{
+	g_signal_emit (view, signals[TITLE_CHANGED], 0, title);
+}
+
+static void
+content_tree_selection_changed_cb (GtkTreeSelection *selection, 
+				   YelpViewContent  *content)
 {
 	YelpViewContentPriv *priv;
  	GtkTreeIter          iter;
@@ -179,6 +239,162 @@ yvc_tree_selection_changed_cb (GtkTreeSelection *selection,
 	}
 }
 
+static void
+content_reader_data_cb (YelpReader      *reader,
+			const gchar     *data,
+			gint             len,
+			YelpViewContent *view)
+{
+	YelpViewContentPriv *priv;
+	
+	g_return_if_fail (YELP_IS_READER (reader));
+	g_return_if_fail (YELP_IS_VIEW_CONTENT (view));
+	
+	priv = view->priv;
+
+	if (priv->first) {
+		yelp_html_clear (priv->html_view);
+		priv->first = FALSE;
+	}
+
+	if (len == -1) {
+		len = strlen (data);
+	}
+
+	if (len <= 0) {
+		return;
+	}
+
+	yelp_html_write (priv->html_view, data, len);
+}
+
+static void
+content_reader_finished_cb (YelpReader      *reader, 
+			    YelpURI         *uri,
+			    YelpViewContent *view)
+{
+	YelpViewContentPriv *priv;
+	GtkTreePath         *path = NULL;
+	
+	g_return_if_fail (YELP_IS_READER (reader));
+	g_return_if_fail (YELP_IS_VIEW_CONTENT (view));
+
+	priv = view->priv;
+
+	if (!priv->first) {
+		yelp_html_close (priv->html_view);
+	}
+
+	path = content_find_path_from_uri (GTK_TREE_MODEL (priv->tree_store),
+					   uri);
+	
+	if (path) {
+		GtkTreeSelection *selection;
+		GtkTreePath      *parent;
+
+		/* Open the correct node in the tree */
+
+		d(g_print ("Found path\n"));
+		
+		selection = gtk_tree_view_get_selection (
+			GTK_TREE_VIEW (priv->content_tree));
+			
+ 		parent = gtk_tree_path_copy (path);
+		
+		gtk_tree_path_up (parent);
+
+		gtk_tree_view_expand_row (GTK_TREE_VIEW (priv->content_tree),
+ 					  parent, FALSE);
+
+		gtk_tree_view_expand_row (GTK_TREE_VIEW (priv->content_tree),
+ 					  path, TRUE);
+
+		parent = gtk_tree_path_copy (path);
+		
+		g_signal_handlers_block_by_func (
+			selection,
+			content_tree_selection_changed_cb,
+			view);
+ 
+		gtk_tree_selection_select_path (selection, path);
+		
+		g_signal_handlers_unblock_by_func (
+			selection,
+			content_tree_selection_changed_cb,
+			view);
+
+		while (gtk_tree_path_up (path)) {
+			gtk_tree_view_expand_row (
+				GTK_TREE_VIEW (priv->content_tree),
+				path, FALSE);
+		}
+
+		gtk_tree_path_free (path);
+		gtk_tree_path_free (parent);
+	}
+
+	gdk_window_set_cursor (priv->html_widget->window, NULL);
+	gtk_widget_grab_focus (priv->html_widget);
+}
+
+static void
+content_reader_error_cb (YelpReader      *reader,
+			 GError          *error, 
+			 YelpViewContent *view)
+{
+	YelpViewContentPriv *priv;
+	
+	g_return_if_fail (YELP_IS_READER (reader));
+	g_return_if_fail (YELP_IS_VIEW_CONTENT (view));
+
+	priv = view->priv;
+
+	/* Popup window */
+	
+	g_warning ("%s\n", error->message);
+}
+
+static GtkTreePath *found_path;
+
+static gboolean
+content_tree_model_foreach (GtkTreeModel *model,
+			    GtkTreePath  *path,
+			    GtkTreeIter  *iter,
+			    YelpURI      *uri)
+{
+	YelpSection *section = NULL;
+	
+	g_return_val_if_fail (GTK_IS_TREE_MODEL (model), TRUE);
+	g_return_val_if_fail (uri != NULL, TRUE);
+	
+	gtk_tree_model_get (model, iter, 
+			    1, &section, 
+			    -1);
+
+	if (!section) {
+		return FALSE;
+	}
+	
+	if (yelp_uri_equal (uri, section->uri)) {
+		found_path = gtk_tree_path_copy (path);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static GtkTreePath *
+content_find_path_from_uri (GtkTreeModel *model, YelpURI *uri)
+{
+	found_path = NULL;
+	
+	gtk_tree_model_foreach (model, 
+				(GtkTreeModelForeachFunc) content_tree_model_foreach,
+				uri);
+
+	return found_path;
+}
+
 GtkWidget *
 yelp_view_content_new (GNode *doc_tree)
 {
@@ -209,7 +425,7 @@ yelp_view_content_new (GNode *doc_tree)
 		GTK_TREE_VIEW (priv->content_tree));
 
 	g_signal_connect (selection, "changed",
-			  G_CALLBACK (yvc_tree_selection_changed_cb), 
+			  G_CALLBACK (content_tree_selection_changed_cb), 
 			  view);
 
         gtk_container_add (GTK_CONTAINER (priv->tree_sw), priv->content_tree);
@@ -223,13 +439,13 @@ yelp_view_content_new (GNode *doc_tree)
 	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
 	
 	gtk_container_add (GTK_CONTAINER (frame), html_sw);
-        gtk_container_add (GTK_CONTAINER (html_sw), 
-			   yelp_html_get_widget (priv->html_view));
+        gtk_container_add (GTK_CONTAINER (html_sw), priv->html_widget);
+
 
 	/* Add the tree and html view to the paned */
 	gtk_paned_add1 (GTK_PANED (view), priv->tree_sw);
         gtk_paned_add2 (GTK_PANED (view), frame);
-        gtk_paned_set_position (GTK_PANED (view), 250);
+        gtk_paned_set_position (GTK_PANED (view), 175);
 
 	return GTK_WIDGET (view);
 }
@@ -288,6 +504,8 @@ yelp_view_content_show_uri (YelpViewContent  *content,
 {
 	YelpViewContentPriv *priv;
 	GNode               *node;
+	GdkCursor           *cursor;
+	gboolean 	     reset_focus = FALSE;
 	
 	g_return_if_fail (YELP_IS_VIEW_CONTENT (content));
 	g_return_if_fail (uri != NULL);
@@ -295,38 +513,62 @@ yelp_view_content_show_uri (YelpViewContent  *content,
 	priv = content->priv;
 
 	if (yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_XML) {
-		/* ghelp uri-scheme /usr/share/gnome/help... */
-		const gchar *docpath;
 
- 		docpath = yelp_uri_get_path (uri);
-
-		if (docpath && strcmp (docpath, priv->current_docpath)) {
+		if (!priv->current_uri || 
+		    !yelp_uri_equal_path (uri, priv->current_uri)) {
 			/* Try to find it in the scrollkeeper database,
 			   doesn't have to exist here */
-			node = yelp_scrollkeeper_get_toc_tree (docpath);
+			node = yelp_scrollkeeper_get_toc_tree (yelp_uri_get_path (uri));
 			
 			if (node) {
 				gtk_widget_show (priv->tree_sw);
 				yelp_view_content_set_tree (content, node);
 				
 			} else {
+				if (gtk_widget_is_focus (priv->tree_sw)) {
+					reset_focus = TRUE;
+				}
+
 				gtk_widget_hide (priv->tree_sw);
 			}
-			
-			g_free (priv->current_docpath);
-			priv->current_docpath = g_strdup (docpath);
 		}
 	} else {
+		if (gtk_widget_is_focus (priv->tree_sw)) {
+			reset_focus = TRUE;
+		}
+
 		gtk_widget_hide (priv->tree_sw);
+		
 	}
 
-	yelp_html_open_uri (priv->html_view, uri, error);
-}
-
-void
-yelp_view_content_stop (YelpViewContent *content)
-{
-	g_return_if_fail (YELP_IS_VIEW_CONTENT (content));
+	if (reset_focus) {
+		gtk_widget_child_focus (GTK_WIDGET (content), 
+					GTK_DIR_TAB_FORWARD);
+	}
 	
-	yelp_html_cancel_loading (content->priv->html_view);
+	priv->first = TRUE;
+	
+	cursor = gdk_cursor_new (GDK_WATCH);
+		
+	gdk_window_set_cursor (priv->html_widget->window, cursor);
+	gdk_cursor_unref (cursor);
+
+	if (priv->current_uri) {
+		yelp_uri_unref (priv->current_uri);
+	}
+	
+	priv->current_uri = yelp_uri_ref (uri);
+	
+	yelp_html_set_base_uri (priv->html_view, uri);
+		
+	if (!yelp_reader_start (priv->reader, uri)) {
+		gchar     *loading = _("Loading...");
+
+		yelp_html_clear (priv->html_view);
+
+		yelp_html_printf (priv->html_view, 
+				  "<html><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"><title>..</title><body><center>%s</center></body></html>", 
+				  loading);
+		yelp_html_close (priv->html_view);
+	}
 }
